==> libphidget22-1.6.20200214.tar.gz: libphidget22-1.6.20200214/src/device/dataadapterdevice.c <==
/*
 * This file is part of libphidget22
 *
 * Copyright 2015 Phidgets Inc <patrick@phidgets.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see
 * <http://www.gnu.org/licenses/>
 */

#include "phidgetbase.h"
#include "gpp.h"
#include "device/dataadapterdevice.h"

// === Internal Functions === //

//initAfterOpen - sets up the initial state of an object, reading in packets from the device if needed
//				  used during attach initialization - on every attach
static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_initAfterOpen(PhidgetDeviceHandle device) {

	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)device;
	//PhidgetReturnCode ret;
	assert(phid);

	switch (device->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		phid->packetCount = 0;
		break;
	case PHIDUID_ADP_RS485_422:
		phid->packetCount = 0;
		phid->storedPacketLength = 0;
		break;
	default:
		MOS_PANIC("Unexpected device");
	}

	return (EPHIDGET_OK);
}

//dataInput - parses device packets
static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_dataInput(PhidgetDeviceHandle device, uint8_t *buffer, size_t length) {
	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)device;
	PhidgetChannelHandle channel;
	uint8_t lastInputState[DATAADAPTER_MAXINPUTS];
	uint16_t receivedPacketCount;
	PhidgetReturnCode ret;
	int endPacket = 0;
	int packetOverrun = 0;
	int error = 0;
	int i, j;

	assert(phid);
	assert(buffer);

	//Parse device packets - store data locally
	switch (device->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		if (length > 0) {
			switch (buffer[0]) {
			case DATAADAPTER_PACKET_DATA_ERROR:
				error = 1; //we intend to fall through to the next here
			case DATAADAPTER_PACKET_DATA:
				receivedPacketCount = unpack16(&buffer[1]);
				if (phid->packetCount != receivedPacketCount) {
					error = 1;
				}
				phid->packetCount = receivedPacketCount + 1;

				memcpy(phid->lastData, buffer + 3, length - 3);

				if ((channel = getChannel(phid, 0)) != NULL) {
					bridgeSendToChannel(channel, BP_DATAIN, "%*R%u", (int)length - 3, phid->lastData, error);
					PhidgetRelease(&channel);
				}
				return (EPHIDGET_OK);
			case STATE_CHANGE:
				for (j = 0; j < phid->devChannelCnts.numInputs; j++) {
	//				inputState[j] = PUNK_BOOL;
					lastInputState[j] = phid->inputState[j];
					phid->inputState[j] = ((buffer[1] & (1 << j)) != 0);
				}

				for (i = 0; i < phid->devChannelCnts.numInputs; i++) {
					int chIndex = i + phid->devChannelCnts.numDataAdapters;
					if ((channel = getChannel(phid, chIndex)) != NULL) {
						if (phid->inputState[i] != PUNK_BOOL && phid->inputState[i] != lastInputState[i]) {
							bridgeSendToChannel(channel, BP_STATECHANGE, "%d", (int)(phid->inputState[i]));
						}
						PhidgetRelease(&channel);
					}
				}

				return (EPHIDGET_OK);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		}
		MOS_PANIC("Unexpected packet type");
	case PHIDUID_ADP_RS485_422:
		if (length > 0) {
			switch (buffer[0]) {
			case DATAADAPTER_PACKET_DATA_ERROR:
				error = 1;  //we intend to fall through to the next here
			case DATAADAPTER_PACKET_DATA_END:
				endPacket = 1;
			case DATAADAPTER_PACKET_DATA:
				receivedPacketCount = unpack16(&buffer[1]);
				if (phid->packetCount != receivedPacketCount) {
					error = 1;
				}
				phid->packetCount = receivedPacketCount + 1;
				memcpy(phid->lastData, buffer + 3, length - 3);

				phid->lastDataLength = phid->storedPacketLength;
				if (phid->storedPacketLength + length - 3 <= 8192) {
					memcpy(&phid->storedPacket[phid->storedPacketLength], phid->lastData, length - 3);
					phid->storedPacketLength += (uint16_t)length - 3;
					if (endPacket == 0)
						return EPHIDGET_OK;
				}
				else
					packetOverrun = 1;

				if ((channel = getChannel(phid, 0)) != NULL) {
					ret = bridgeSendToChannel(channel, BP_DATAIN, "%*R%u", phid->storedPacketLength, phid->storedPacket, error);
					if (ret != EPHIDGET_NOSPC)
						phid->storedPacketLength = 0;

					PhidgetRelease(&channel);
				}

				if (packetOverrun) {
					memcpy(&phid->storedPacket[phid->storedPacketLength], phid->lastData, length - 3);
					phid->storedPacketLength += (uint16_t)length - 3;
				}

				return (EPHIDGET_OK);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		}
		MOS_PANIC("Unexpected packet type");
	default:
		MOS_PANIC("Unexpected device");
	}

}

static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_bridgeInput(PhidgetChannelHandle ch, BridgePacket *bp) {
	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)ch->parent;
	unsigned char buffer[MAX_OUT_PACKET_SIZE] = { 0 };
	PhidgetReturnCode ret;
	size_t len;

	assert(phid->phid.deviceInfo.class == PHIDCLASS_DATAADAPTER);


	switch (((PhidgetDeviceHandle)phid)->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		switch (ch->class) {
		case PHIDCHCLASS_DATAADAPTER:
			switch (bp->vpkt) {
			case BP_DATAOUT:
				ret = sendData(phid, bp);

				return ret;
			case BP_SETBAUDRATE:
				len = 4;
				pack32(buffer, getBridgePacketUInt32(bp, 0));
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_BAUD_RATE, 0, buffer, &len, 100);
			case BP_SETPARITY:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PARITY_MODE, 0, buffer, &len, 100);
			case BP_SETSTOPBITS:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_STOP_BITS, 0, buffer, &len, 100);
			case BP_SETDATABITS:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketUInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_DATA_BITS, 0, buffer, &len, 100);
			case BP_SETHANDSHAKEMODE:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RTS_CTS_MODE, 0, buffer, &len, 100);
			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RESET, 0, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_ENABLE, 0, buffer, &len, 100);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		case PHIDCHCLASS_DIGITALINPUT:
			switch (bp->vpkt) {
			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_RESET, ch->uniqueIndex, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_ENABLE, ch->uniqueIndex, buffer, &len, 100);
			}
		case PHIDCHCLASS_DIGITALOUTPUT:
			switch (bp->vpkt) {
			case BP_SETDUTYCYCLE:
			{
				double state = getBridgePacketDouble(bp, 0);
				if (state != 0 && state != 1)
					return EPHIDGET_INVALIDARG;
				buffer[0] = (uint8_t)state;
				len = 1;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_DIGITALOUTPUT_SETDUTYCYCLE, ch->uniqueIndex, buffer, &len, 100);
			}

			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_RESET, ch->uniqueIndex, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_ENABLE, ch->uniqueIndex, buffer, &len, 100);
			}
		default:
			MOS_PANIC("Unexpected Channel Class");
		}
	case PHIDUID_ADP_RS485_422:
		switch (bp->vpkt) {
		case BP_DATAOUT:
			ret = sendData(phid, bp);

			return ret;
		case BP_SETBAUDRATE:
			len = 4;
			pack32(buffer, getBridgePacketUInt32(bp, 0));
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_BAUD_RATE, 0, buffer, &len, 100);
		case BP_SETPARITY:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PARITY_MODE, 0, buffer, &len, 100);
		case BP_SETSTOPBITS:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_STOP_BITS, 0, buffer, &len, 100);
		case BP_SETDATABITS:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketUInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_DATA_BITS, 0, buffer, &len, 100);
		case BP_SETHANDSHAKEMODE:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RTS_CTS_MODE, 0, buffer, &len, 100);
		case BP_OPENRESET:
		case BP_CLOSERESET:
			len = 0;
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RESET, 0, buffer, &len, 100);
		case BP_ENABLE:
			len = 0;
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_ENABLE, 0, buffer, &len, 100);

		case BP_SETPROTOCOL:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PROTOCOL, 0, buffer, &len, 100);
		}
	default:
		MOS_PANIC("Unexpected device");
	}
}

static void CCONV
PhidgetDataAdapterDevice_free(PhidgetDeviceHandle *phid) {

	mos_free(*phid, sizeof(struct _PhidgetDataAdapterDevice));
	*phid = NULL;
}

PhidgetReturnCode
PhidgetDataAdapterDevice_create(PhidgetDataAdapterDeviceHandle *phidp) {
	DEVICECREATE_BODY(DataAdapterDevice, PHIDCLASS_DATAADAPTER);
	return (EPHIDGET_OK);
}


PhidgetReturnCode sendData( PhidgetDataAdapterDeviceHandle phid, BridgePacket* bp) {
	size_t len = ((size_t)getBridgePacketInt32(bp, 0));
	PhidgetReturnCode ret;
	size_t packetLen;
	size_t i = 0;

	uint8_t buf[64];

	ret = EPHIDGET_OK;

	if (len > 0) {
		if (len <= 63) {
			packetLen = len + 1;
			buf[0] = 1;
			memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1), len);
			ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);
			if (ret != EPHIDGET_OK)
				return ret;

		} else {
			for (i = 0; i + 63 < len; i += 63) {
				packetLen = 64;
				if (i == 0)
					buf[0] = 1;
				else
					buf[0] = 0;
				memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1)+i, packetLen - 1);
				ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);

				if(ret != EPHIDGET_OK)
					return ret;
			}
			if (i != len) {
				packetLen = len - i + 1;
				buf[0] = 0;
				memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1) + i, packetLen - 1);
				ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);
				if (ret != EPHIDGET_OK)
					return ret;
			}
		}
	}
	return ret;
}

==> libphidget22-1.6.20200306.tar.gz: libphidget22-1.6.20200306/src/device/dataadapterdevice.c <==
/*
 * This file is part of libphidget22
 *
 * Copyright 2015 Phidgets Inc <patrick@phidgets.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see
 * <http://www.gnu.org/licenses/>
 */

#include "phidgetbase.h"
#include "gpp.h"
#include "device/dataadapterdevice.h"

// === Internal Functions === //

//initAfterOpen - sets up the initial state of an object, reading in packets from the device if needed
//				  used during attach initialization - on every attach
static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_initAfterOpen(PhidgetDeviceHandle device) {

	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)device;
	//PhidgetReturnCode ret;
	assert(phid);

	switch (device->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		phid->packetCount = 0;
		break;
	case PHIDUID_ADP_RS485_422:
		phid->packetCount = 0;
		phid->storedPacketLength = 0;
		break;
	default:
		MOS_PANIC("Unexpected device");
	}

	return (EPHIDGET_OK);
}

//dataInput - parses device packets
static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_dataInput(PhidgetDeviceHandle device, uint8_t *buffer, size_t length) {
	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)device;
	PhidgetChannelHandle channel;
	uint8_t lastInputState[DATAADAPTER_MAXINPUTS];
	uint16_t receivedPacketCount;
	PhidgetReturnCode ret;
	int endPacket = 0;
	int packetOverrun = 0;
	int error = 0;
	int i, j;

	assert(phid);
	assert(buffer);

	//Parse device packets - store data locally
	switch (device->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		if (length > 0) {
			switch (buffer[0]) {
			case DATAADAPTER_PACKET_DATA_ERROR:
				error = 1; //we intend to fall through to the next here
			case DATAADAPTER_PACKET_DATA:
				receivedPacketCount = unpack16(&buffer[1]);
				if (phid->packetCount != receivedPacketCount) {
					error = 1;
				}
				phid->packetCount = receivedPacketCount + 1;

				memcpy(phid->lastData, buffer + 3, length - 3);

				if ((channel = getChannel(phid, 0)) != NULL) {
					bridgeSendToChannel(channel, BP_DATAIN, "%*R%u", (int)(length - 3), phid->lastData, error);
					PhidgetRelease(&channel);
				}
				return (EPHIDGET_OK);
			case STATE_CHANGE:
				for (j = 0; j < phid->devChannelCnts.numInputs; j++) {
	//				inputState[j] = PUNK_BOOL;
					lastInputState[j] = phid->inputState[j];
					phid->inputState[j] = ((buffer[1] & (1 << j)) != 0);
				}

				for (i = 0; i < phid->devChannelCnts.numInputs; i++) {
					int chIndex = i + phid->devChannelCnts.numDataAdapters;
					if ((channel = getChannel(phid, chIndex)) != NULL) {
						if (phid->inputState[i] != PUNK_BOOL && phid->inputState[i] != lastInputState[i]) {
							bridgeSendToChannel(channel, BP_STATECHANGE, "%d", (int)(phid->inputState[i]));
						}
						PhidgetRelease(&channel);
					}
				}

				return (EPHIDGET_OK);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		}
		MOS_PANIC("Unexpected packet type");
	case PHIDUID_ADP_RS485_422:
		if (length > 0) {
			switch (buffer[0]) {
			case DATAADAPTER_PACKET_DATA_ERROR:
				error = 1;  //we intend to fall through to the next here
			case DATAADAPTER_PACKET_DATA_END:
				endPacket = 1;
			case DATAADAPTER_PACKET_DATA:
				receivedPacketCount = unpack16(&buffer[1]);
				if (phid->packetCount != receivedPacketCount) {
					error = 1;
				}
				phid->packetCount = receivedPacketCount + 1;
				memcpy(phid->lastData, buffer + 3, length - 3);

				phid->lastDataLength = phid->storedPacketLength;
				if (phid->storedPacketLength + length - 3 <= 8192) {
					memcpy(&phid->storedPacket[phid->storedPacketLength], phid->lastData, length - 3);
					phid->storedPacketLength += (uint16_t)length - 3;
					if (endPacket == 0)
						return EPHIDGET_OK;
				}
				else
					packetOverrun = 1;

				if ((channel = getChannel(phid, 0)) != NULL) {
					ret = bridgeSendToChannel(channel, BP_DATAIN, "%*R%u", (int)phid->storedPacketLength, phid->storedPacket, error);
					if (ret != EPHIDGET_NOSPC)
						phid->storedPacketLength = 0;

					PhidgetRelease(&channel);
				}

				if (packetOverrun) {
					memcpy(&phid->storedPacket[phid->storedPacketLength], phid->lastData, length - 3);
					phid->storedPacketLength += (uint16_t)length - 3;
				}

				return (EPHIDGET_OK);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		}
		MOS_PANIC("Unexpected packet type");
	default:
		MOS_PANIC("Unexpected device");
	}

}

static PhidgetReturnCode CCONV
PhidgetDataAdapterDevice_bridgeInput(PhidgetChannelHandle ch, BridgePacket *bp) {
	PhidgetDataAdapterDeviceHandle phid = (PhidgetDataAdapterDeviceHandle)ch->parent;
	unsigned char buffer[MAX_OUT_PACKET_SIZE] = { 0 };
	PhidgetReturnCode ret;
	size_t len;

	assert(phid->phid.deviceInfo.class == PHIDCLASS_DATAADAPTER);


	switch (((PhidgetDeviceHandle)phid)->deviceInfo.UDD->uid) {
	case PHIDUID_ADP_RS232:
		switch (ch->class) {
		case PHIDCHCLASS_DATAADAPTER:
			switch (bp->vpkt) {
			case BP_DATAOUT:
				ret = sendData(phid, bp);

				return ret;
			case BP_SETBAUDRATE:
				len = 4;
				pack32(buffer, getBridgePacketUInt32(bp, 0));
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_BAUD_RATE, 0, buffer, &len, 100);
			case BP_SETPARITY:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PARITY_MODE, 0, buffer, &len, 100);
			case BP_SETSTOPBITS:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_STOP_BITS, 0, buffer, &len, 100);
			case BP_SETDATABITS:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketUInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_DATA_BITS, 0, buffer, &len, 100);
			case BP_SETHANDSHAKEMODE:
				len = 1;
				buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RTS_CTS_MODE, 0, buffer, &len, 100);
			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RESET, 0, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_ENABLE, 0, buffer, &len, 100);
			default:
				MOS_PANIC("Unexpected packet type");
			}
		case PHIDCHCLASS_DIGITALINPUT:
			switch (bp->vpkt) {
			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_RESET, ch->uniqueIndex, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_ENABLE, ch->uniqueIndex, buffer, &len, 100);
			}
		case PHIDCHCLASS_DIGITALOUTPUT:
			switch (bp->vpkt) {
			case BP_SETDUTYCYCLE:
			{
				double state = getBridgePacketDouble(bp, 0);
				if (state != 0 && state != 1)
					return EPHIDGET_INVALIDARG;
				buffer[0] = (uint8_t)state;
				len = 1;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_DIGITALOUTPUT_SETDUTYCYCLE, ch->uniqueIndex, buffer, &len, 100);
			}

			case BP_OPENRESET:
			case BP_CLOSERESET:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_RESET, ch->uniqueIndex, buffer, &len, 100);
			case BP_ENABLE:
				len = 0;
				return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_CHANNEL_WRITE, VINT_PACKET_TYPE_PHIDGET_ENABLE, ch->uniqueIndex, buffer, &len, 100);
			}
		default:
			MOS_PANIC("Unexpected Channel Class");
		}
	case PHIDUID_ADP_RS485_422:
		switch (bp->vpkt) {
		case BP_DATAOUT:
			ret = sendData(phid, bp);

			return ret;
		case BP_SETBAUDRATE:
			len = 4;
			pack32(buffer, getBridgePacketUInt32(bp, 0));
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_BAUD_RATE, 0, buffer, &len, 100);
		case BP_SETPARITY:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PARITY_MODE, 0, buffer, &len, 100);
		case BP_SETSTOPBITS:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_STOP_BITS, 0, buffer, &len, 100);
		case BP_SETDATABITS:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketUInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_DATA_BITS, 0, buffer, &len, 100);
		case BP_SETHANDSHAKEMODE:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RTS_CTS_MODE, 0, buffer, &len, 100);
		case BP_OPENRESET:
		case BP_CLOSERESET:
			len = 0;
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_RESET, 0, buffer, &len, 100);
		case BP_ENABLE:
			len = 0;
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_ENABLE, 0, buffer, &len, 100);

		case BP_SETPROTOCOL:
			len = 1;
			buffer[0] = (uint8_t)getBridgePacketInt32(bp, 0);
			return PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_DEVICE_WRITE, DATAADAPTER_PROTOCOL, 0, buffer, &len, 100);
		}
	default:
		MOS_PANIC("Unexpected device");
	}
}

static void CCONV
PhidgetDataAdapterDevice_free(PhidgetDeviceHandle *phid) {

	mos_free(*phid, sizeof(struct _PhidgetDataAdapterDevice));
	*phid = NULL;
}

PhidgetReturnCode
PhidgetDataAdapterDevice_create(PhidgetDataAdapterDeviceHandle *phidp) {
	DEVICECREATE_BODY(DataAdapterDevice, PHIDCLASS_DATAADAPTER);
	return (EPHIDGET_OK);
}


PhidgetReturnCode sendData( PhidgetDataAdapterDeviceHandle phid, BridgePacket* bp) {
	size_t len = ((size_t)getBridgePacketInt32(bp, 0));
	PhidgetReturnCode ret;
	size_t packetLen;
	size_t i = 0;

	uint8_t buf[64];

	ret = EPHIDGET_OK;

	if (len > 0) {
		if (len <= 63) {
			packetLen = len + 1;
			buf[0] = 1;
			memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1), len);
			ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);
			if (ret != EPHIDGET_OK)
				return ret;

		} else {
			for (i = 0; i + 63 < len; i += 63) {
				packetLen = 64;
				if (i == 0)
					buf[0] = 1;
				else
					buf[0] = 0;
				memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1)+i, packetLen - 1);
				ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);

				if(ret != EPHIDGET_OK)
					return ret;
			}
			if (i != len) {
				packetLen = len - i + 1;
				buf[0] = 0;
				memcpy(&buf[1], (const uint8_t *)getBridgePacketPtr(bp, 1) + i, packetLen - 1);
				ret = PhidgetDevice_transferpacket(bp->iop, (PhidgetDeviceHandle)phid, PHIDGETUSB_REQ_BULK_WRITE, 0, 0, buf, &packetLen, 5000);
				if (ret != EPHIDGET_OK)
					return ret;
			}
		}
	}
	return ret;
}
